@@ -3,16 +3,16 @@
 # because we instead use JSON messages to transmit any changes between the
 # Python and Javascript parts so that we can re-use this for the Qt client.
 
-from astropy.time import Time
 import ipywidgets as widgets
 import numpy as np
-from traitlets import Unicode, Float, default, link, directional_link
+from traitlets import Unicode, default, link, directional_link
 
 from ipyevents import Event as DOMListener
 from ipykernel.comm import Comm
 
 from .core import BaseWWTWidget
 from .layers import ImageLayer
+from .logger import logger
 from .jupyter_server import serve_file
 
 __all__ = ['WWTJupyterWidget', 'WWTLabApplication', 'connect_to_app']
@@ -31,6 +31,13 @@ dom_listener = DOMListener()
 class WWTJupyterWidget(widgets.DOMWidget, BaseWWTWidget):
     """
     An AAS WorldWide Telescope Jupyter widget.
+
+    Parameters
+    ----------
+
+    hide_all_chrome : optional `bool`
+        Configures the WWT frontend to hide all user-interface "chrome".
+
     """
 
     _view_name = Unicode('WWTView').tag(sync=True)
@@ -40,52 +47,60 @@ class WWTJupyterWidget(widgets.DOMWidget, BaseWWTWidget):
     _view_module_version = Unicode(VIEW_MODULE_VERSION).tag(sync=True)
     _model_module_version = Unicode(MODEL_MODULE_VERSION).tag(sync=True)
 
-    # wwt=None tag needed to avoid linkage to 'wwt.settings.set_' type traits
-    # (see _on_trait_change() in core.py)
-    _ra = Float(0.0).tag(sync=True, wwt=None)
-    _dec = Float(0.0).tag(sync=True, wwt=None)
-    _fov = Float(60.0).tag(sync=True, wwt=None)
-    # this is the WWT engine's clock at last check-in:
-    _datetime = Unicode('2017-03-09T12:30:00').tag(sync=True, wwt=None)
-    # this is system clock at last check-in:
-    _systemDatetime = Unicode('2017-03-09T12:30:00').tag(sync=True, wwt=None)
-    # this is the rate at which the WWT engine clock proceeds relative to the system clock:
-    _timeRate = Float(1.0).tag(sync=True, wwt=None)
+    _appUrl = Unicode('').tag(sync=True)
+
+    def __init__(self, hide_all_chrome=False):
+        # In the future we might want to make it possible to use the WWT-hosted
+        # app instead of the bundled version.
+        #
+        # The JS frontend will automagically prepend the Jupyter base URL if
+        # needed and turn this into an absolute URL.
+
+        _maybe_perpetrate_mega_kernel_hack()
+
+        self._appUrl = '/wwt/research/'
 
-    def __init__(self):
         widgets.DOMWidget.__init__(self)
-        BaseWWTWidget.__init__(self)
         dom_listener.source = self
         dom_listener.prevent_default_action = True
         dom_listener.watched_events = ['wheel']
+
         self._controls = None
 
+        self.on_msg(self._on_ipywidgets_message)
+
+        BaseWWTWidget.__init__(self, hide_all_chrome=hide_all_chrome)
+
+    def _on_ipywidgets_message(self, widget, content, buffers):
+        """
+        Called when we receive a "custom" ipywidgets message.
+
+        NOTE: because this code is run asynchronously in Jupyter's comms
+        architecture, exceptions and printouts don't get reported to the user --
+        they just disappear. I don't know if there's a "right" way to address
+        that.
+        """
+
+        # Special message from the ipywidgets bridge to indicate
+        # when the first widget view is ready to accept messages.
+        if content.get('type') == 'wwt_jupyter_widget_status':
+            self._on_app_status_change(alive=content['alive'])
+
+        self._on_app_message_received(content)
+
+    def _actually_send_msg(self, payload):
+        """
+        Send a message to the app. In ipywidgets this is easy.
+        """
+        self.send(payload)
+
     @default('layout')
     def _default_layout(self):
         return widgets.Layout(height='400px', align_self='stretch')
 
-    def _send_msg(self, **kwargs):
-        self.send(kwargs)
-
     def _serve_file(self, filename, extension=''):
         return serve_file(filename, extension=extension)
 
-    def _get_view_data(self, field):
-        if field == 'ra':
-            return self._ra
-        elif field == 'dec':
-            return self._dec
-        elif field == 'fov':
-            return self._fov
-        elif field == 'datetime':
-            engine_time = Time(self._datetime, format='isot')
-            system_time = Time(self._systemDatetime, format='isot')
-            engine_delta = self._timeRate * (Time.now() - system_time)
-            return engine_time + engine_delta
-        else:
-            raise ValueError("'field' should be one of: 'ra', 'dec', "
-                             "'fov', or 'datetime'")
-
     def _create_image_layer(self, **kwargs):
         """Returns a specialized subclass of ImageLayer that has some extra hooks for
         creating UI control points.
@@ -212,63 +227,45 @@ class WWTLabApplication(BaseWWTWidget):
     one specific notebook. The Python API is the same, it's just that the JSON
     messages we send are routed to the separate application rather than our own
     iframe.
-
     """
+
     _comm = None
     _controls = None
 
-    # View state that gets synchronized back to us. This is the same scheme as
-    # the widget, just with manual synchronization over our comm to the viewer
-    # app.
-    _raRad = 0.0
-    _decRad = 0.0
-    _fovDeg = 60.0
-    _engineTime = Time('2017-03-09T12:30:00', format='isot')
-    _systemTime = Time('2017-03-09T12:30:00', format='isot')
-    _timeRate = 1.0
-
     def __init__(self):
+        _maybe_perpetrate_mega_kernel_hack()
+
         self._comm = Comm(target_name='@wwtelescope/jupyterlab:research', data={})
-        self._comm.on_msg(self._on_message_received)
+        self._comm.on_msg(self._on_comm_message_received)
         self._comm.open()
-        self._send_msg(event='trigger')  # get bidirectional updates flowing
 
-        BaseWWTWidget.__init__(self)
+        super(WWTLabApplication, self).__init__()
 
-    def _send_msg(self, **kwargs):
-        self._comm.send(kwargs)
+    def _on_comm_message_received(self, msg):
+        """
+        Called when we receive a comms message.
 
-    def _on_message_received(self, msg):
+        NOTE: because this code is run asynchronously in Jupyter's comms
+        architecture, exceptions and printouts don't get reported to the user --
+        they just disappear. I don't know if there's a "right" way to address
+        that.
+        """
         payload = msg['content']['data']
-        if payload['type'] != 'wwt_view_state':
-            return
 
-        try:
-            self._raRad = float(payload['raRad'])
-            self._decRad = float(payload['decRad'])
-            self._fovDeg = float(payload['fovDeg'])
-            self._engineTime = Time(payload['engineClockISOT'], format='isot')
-            self._systemTime = Time(payload['systemClockISOT'], format='isot')
-            self._timeRate = float(payload['engineClockRateFactor'])
-        except ValueError:
-            pass  # report a warning somehow?
+        # Special message from the hub indicating app liveness status
+        if payload.get('type') == 'wwt_jupyter_viewer_status':
+            self._on_app_status_change(alive=payload['alive'])
+            # don't return -- maybe someone downstream can use this, and message
+            # processing needs to handle all sorts of unexpected messages anyway
+
+        self._on_app_message_received(payload)
+
+    def _actually_send_msg(self, payload):
+        self._comm.send(payload)
 
     def _serve_file(self, filename, extension=''):
         return serve_file(filename, extension=extension)
 
-    def _get_view_data(self, field):
-        if field == 'ra':
-            return self._raRad * R2H
-        elif field == 'dec':
-            return self._decRad * R2D
-        elif field == 'fov':
-            return self._fovDeg
-        elif field == 'datetime':
-            engine_delta = self._timeRate * (Time.now() - self._systemTime)
-            return self._engineTime + engine_delta
-        else:
-            raise ValueError('internal problem: unexpected "field" value')
-
     def _create_image_layer(self, **kwargs):
         """Returns a specialized subclass of ImageLayer that has some extra hooks for
         creating UI control points.
@@ -313,3 +310,148 @@ def connect_to_app():
     # have the user call a function with this name, than to create a "connection
     # object".
     return WWTLabApplication()
+
+
+def _maybe_perpetrate_mega_kernel_hack():
+    """
+    OK. So.
+
+    The Python code running in this process is communicating with the WWT
+    JavaScript frontend, which is running in another process that is possibly on
+    the other side of the internet. Various WWT operations such as loading
+    imagery need that frontend to perform web requests and the like. So much of
+    what we want to do is just inherently asynchronous.
+
+    Which is OK! We require sufficiently new Python that we can rely on
+    async/await being available, and the Jupyter infrastructure is extremely
+    async-friendly.
+
+    Except that we have a problem. When our process is running inside Jupyter,
+    all code executions are inherently triggered by "shell messages" telling the
+    kernel to run some code. Shell messages are also used for Jupyter's "comms"
+    infrastructure, which is how kernels get updates from the WWT frontend(s)
+    about things that are going on. The problem is that the Python Jupyter
+    kernels only process shell messages *sequentially*: while message evaluation
+    can be asynchronous, they can't be evaluated simultaneously. So if we've got
+    asynchronous user code that's attempting to do stuff with pywwt, we can't
+    receive any updates from WWT while that code is running. This fundamentally
+    breaks our ability to allow users to write code that interacts with WWT
+    asynchronously.
+
+    Our mega-hack solution is to allow some shell messages to be marked for
+    "expedited" processing. We hack the kernel so that such messages can be
+    processed even while another shell message -- such as an execution request
+    -- is being dealt with. This breaks the logjam.
+
+    Expedited messages have the following structure:
+
+    {
+        'content': {
+            '_pywwtExpedite': true
+        }
+    }
+
+    This allows ipywidgets custom messages to be marked for expedited
+    processing, which we need for our ipywidget support.
+
+    To amp up the debugging spew so that you can see what's going on inside
+    the kernel (which is captured in the Jupyter Server output), run:
+
+    ```
+    import ipykernel.kernelbase
+    ipykernel.kernelbase.Kernel.instance().log.setLevel('DEBUG')
+    ```
+
+    Please forgive me.
+    """
+
+    try:
+        _maybe_perpetrate_mega_kernel_hack_inner()
+    except Exception:
+        logger.exception('failed to set up Jupyter kernel async hack')
+
+
+def _maybe_perpetrate_mega_kernel_hack_inner():
+    import asyncio
+    import inspect
+    import ipykernel.kernelbase
+
+    kernel = ipykernel.kernelbase.Kernel.instance()
+    orig_schedule_dispatch = kernel.schedule_dispatch
+
+    if getattr(kernel, '_pywwt_mega_hack_installed', False):
+        return
+
+    # If asyncio doesn't think that there's a running event loop, we don't seem
+    # to be running as a client of a full-on Jupyter server. In which case, play
+    # it safe and don't do anything.
+
+    try:
+        asyncio.get_running_loop()
+    except RuntimeError:
+        return
+
+    # OK, it looks like we should try to do this ...
+
+    async def dispatch_one_expedited_shell_message(idents, msg):
+        """
+        A bastardized version of `kernel.dispatch_shell`. Our arguments will be
+        the result of `kernel.session.feed_identities`.
+
+        Expedited shell messages have to be processed as straightforwardly as
+        possible since their processing can happen while other shell message
+        processing is happening. The kernel code is *not* build to handle the
+        recursive-y, race-y things that can happen in this situation. So we
+        avoid all function calls that we can possible get away with skipping.
+        No set_parent, no should_handle, etc.
+        """
+
+        try:
+            msg = kernel.session.deserialize(msg, content=True, copy=False)
+        except Exception:
+            kernel.log.error("Invalid Expedited Message (pywwt)", exc_info=True)
+            return
+
+        msg_type = msg['header']['msg_type']
+        handler = kernel.shell_handlers.get(msg_type, None)
+
+        if handler is None:
+            kernel.log.warning("Unknown expedited message type (pywwt): %r", msg_type)
+        else:
+            kernel.log.debug("expedited (pywwt) %s: %s", msg_type, msg)
+
+            try:
+                result = handler(kernel.shell_stream, idents, msg)
+                if inspect.isawaitable(result):
+                    kernel.log.error("Expedited message (pywwt) produce an awaitable result")
+            except Exception:
+                kernel.log.error("Exception in expedited message handler (pywwt):", exc_info=True)
+
+    def pywwt_schedule_shell_dispatch_with_expedite(*args):
+        """
+        A replacement of `kernel.schedule_dispatch` for shell messages.
+
+        This peeks inside the message and triggers expedited processing if
+        called for. If there are any issues, we fall back to regular processing.
+
+        """
+        expedited_it = False
+
+        try:
+            (msg,) = args
+            idents, msg = kernel.session.feed_identities(msg, copy=False)
+            # We can't deserialize() here: each message can only be deserialized once
+            # due to the replay prevention framework.
+            peek_content = kernel.session.unpack(msg[4].bytes)
+            expedite_flag = peek_content.get('data', {}).get('content', {}).get('_pywwtExpedite')
+
+            if expedite_flag:
+                kernel.io_loop.add_callback(dispatch_one_expedited_shell_message, idents, msg)
+                expedited_it = True
+        finally:
+            if not expedited_it:
+                orig_schedule_dispatch(kernel.dispatch_shell, *args)
+
+    kernel.shell_stream.on_recv(pywwt_schedule_shell_dispatch_with_expedite, copy=False)
+    kernel._pywwt_mega_hack_installed = True
+    logger.debug('installed Jupyter kernel message expedite hack')
